/*****************************************************************************
*   $Id: eiffel.c,v 1.9 1998/10/18 18:23:42 darren Exp $
*
*   Copyright (c) 1998, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains functions for generating tags for Eiffel language
*   files.
*****************************************************************************/

/*============================================================================
=   Include files
============================================================================*/
#include "general.h"

#include <string.h>
#include <limits.h>
#include <ctype.h>	/* to define tolower() */

#include "debug.h"
#include "eiffel.h"
#include "entry.h"
#include "options.h"
#include "read.h"
#include "vstring.h"

/*============================================================================
=   Macros
============================================================================*/
#define hashIndex(c)		((c) - 'a')
#define activeTag(st)		((st)->tag[(int)(st)->buf1])
#define activeName(st)		(activeTag(st).name)
#define swapNameBuffers(st)	((st)->buf1 = (boolean)!(st)->buf1)
#define isFreeOperatorChar(c)	((c) == '@' || (c) == '#' || \
				 (c) == '|' || (c) == '&')

#define isExternCDecl(st,c)	((c) == STRING_SYMBOL  &&  !(st)->gotName  && \
				 (st)->scope == SCOPE_EXTERN)

/*============================================================================
=   Data declarations
============================================================================*/

/*  Used to specify type of keyword.
 */
typedef enum _keywordId {
    KEYWORD_NONE,
    KEYWORD_alias, KEYWORD_all, KEYWORD_and, KEYWORD_as, KEYWORD_check,
    KEYWORD_class, KEYWORD_creation, KEYWORD_Current, KEYWORD_debug,
    KEYWORD_deferred, KEYWORD_do, KEYWORD_else, KEYWORD_elseif, KEYWORD_end,
    KEYWORD_ensure, KEYWORD_expanded, KEYWORD_export, KEYWORD_external,
    KEYWORD_false, KEYWORD_feature, KEYWORD_from, KEYWORD_frozen, KEYWORD_if,
    KEYWORD_implies, KEYWORD_indexing, KEYWORD_infix, KEYWORD_inherit,
    KEYWORD_inspect, KEYWORD_invariant, KEYWORD_is, KEYWORD_like,
    KEYWORD_local, KEYWORD_loop, KEYWORD_not, KEYWORD_obsolete, KEYWORD_old,
    KEYWORD_once, KEYWORD_or, KEYWORD_prefix, KEYWORD_redefine, KEYWORD_rename,
    KEYWORD_require, KEYWORD_rescue, KEYWORD_Result, KEYWORD_retry,
    KEYWORD_select, KEYWORD_separate, KEYWORD_strip, KEYWORD_then,
    KEYWORD_true, KEYWORD_undefine, KEYWORD_unique, KEYWORD_until,
    KEYWORD_variant, KEYWORD_when, KEYWORD_xor
} keywordId;

/*  Used to determine whether keyword is valid for the token language and
 *  what its ID is.
 */
typedef struct _keywordDesc {
    const char *name;
    keywordId id;
} keywordDesc;

typedef enum _tokenType {
    TOKEN_UNDEFINED,
    TOKEN_BANG,
    TOKEN_CHARACTER,
    TOKEN_CLOSE_BRACE,
    TOKEN_CLOSE_BRACKET,
    TOKEN_CLOSE_PAREN,
    TOKEN_COMMA,
    TOKEN_COLON,
    TOKEN_DOT,
    TOKEN_DOLLAR,
    TOKEN_EOF,
    TOKEN_IDENTIFIER,
    TOKEN_KEYWORD,
    TOKEN_NUMERIC,
    TOKEN_OPEN_BRACE,
    TOKEN_OPEN_BRACKET,
    TOKEN_OPEN_PAREN,
    TOKEN_OPERATOR,
    TOKEN_OTHER,
    TOKEN_SEPARATOR,
    TOKEN_STRING
} tokenType;

/*  Describes the statement tokenly undergoing analysis.
 */
typedef struct _statementInfo {
    tagScope	scope;
} statementInfo;

typedef struct _tokenInfo {
    tokenType		type;
    keywordId		keyword;
    vString *		string;
} tokenInfo;

/*============================================================================
=   Data definitions
============================================================================*/

enum { HashSize = ('z' - 'a' + 1) };	/* 'a' through 'z' */
static short EiffelKeywordHash[(int)HashSize];

static const keywordDesc EiffelKeywordTable[] = {
    /* keyword		keyword ID */
    { "alias",		KEYWORD_alias		},
    { "all",		KEYWORD_all		},
    { "and",		KEYWORD_and		},
    { "as",		KEYWORD_as		},
    { "check",		KEYWORD_check		},
    { "class",		KEYWORD_class		},
    { "creation",	KEYWORD_creation	},
    { "Current",	KEYWORD_Current		},
    { "debug",		KEYWORD_debug		},
    { "deferred",	KEYWORD_deferred	},
    { "do",		KEYWORD_do		},
    { "else",		KEYWORD_else		},
    { "elseif",		KEYWORD_elseif		},
    { "end",		KEYWORD_end		},
    { "ensure",		KEYWORD_ensure		},
    { "expanded",	KEYWORD_expanded	},
    { "export",		KEYWORD_export		},
    { "external",	KEYWORD_external	},
    { "false",		KEYWORD_false		},
    { "feature",	KEYWORD_feature		},
    { "from",		KEYWORD_from		},
    { "frozen",		KEYWORD_frozen		},
    { "if",		KEYWORD_if		},
    { "implies",	KEYWORD_implies		},
    { "indexing",	KEYWORD_indexing	},
    { "infix",		KEYWORD_infix		},
    { "inherit",	KEYWORD_inherit		},
    { "inspect",	KEYWORD_inspect		},
    { "invariant",	KEYWORD_invariant	},
    { "is",		KEYWORD_is		},
    { "like",		KEYWORD_like		},
    { "local",		KEYWORD_local		},
    { "loop",		KEYWORD_loop		},
    { "not",		KEYWORD_not		},
    { "obsolete",	KEYWORD_obsolete	},
    { "old",		KEYWORD_old		},
    { "once",		KEYWORD_or		},
    { "prefix",		KEYWORD_prefix		},
    { "redefine",	KEYWORD_redefine	},
    { "rename",		KEYWORD_rename		},
    { "require",	KEYWORD_require		},
    { "rescue",		KEYWORD_rescue		},
    { "Result",		KEYWORD_Result		},
    { "retry",		KEYWORD_retry		},
    { "select",		KEYWORD_select		},
    { "separate",	KEYWORD_separate	},
    { "strip",		KEYWORD_strip		},
    { "then",		KEYWORD_then		},
    { "true",		KEYWORD_true		},
    { "undefine",	KEYWORD_undefine	},
    { "unique",		KEYWORD_unique		},
    { "until",		KEYWORD_until		},
    { "variant",	KEYWORD_variant		},
    { "when",		KEYWORD_when		},
    { "xor",		KEYWORD_xor		}
};

static const size_t EiffelKeywordTableSize =
		sizeof(EiffelKeywordTable)/sizeof(EiffelKeywordTable[0]);

/*============================================================================
=   Function prototypes
============================================================================*/
static boolean vStringEquiv __ARGS((vString *const string, const char *const s));
static void qualifyClassTag __ARGS((const char *const name, memberInfo *const pMember));
static void qualifyFeatureTag __ARGS((const char *const name, memberInfo *const pMember, const boolean isExported));
static void qualifyLocalTag __ARGS((const char *const name, memberInfo *const pMember));
static int skipToCharacter __ARGS((const int c));
static const char *parseInteger __ARGS((int c));
static const char *parseNumeric __ARGS((int c));
static int parseEscapedCharacter __ARGS((void));
static int parseCharacter __ARGS((void));
static void parseString __ARGS((vString *const string));
static void parseIdentifier __ARGS((vString *const string, const int firstChar));
static void parseFreeOperator __ARGS((vString *const string, const int firstChar));
static keywordId analyzeToken __ARGS((const char *const name));
static boolean parseToken __ARGS((tokenInfo *const token));
static boolean isIdentifierMatch __ARGS((const tokenInfo *const token, const char *const name));
static boolean locateToken __ARGS((const tokenType type, const boolean skipPast, tokenInfo *const token));
static boolean locateKeyword __ARGS((const keywordId keyword, const boolean skipPast, tokenInfo *const token));
static boolean parseLocal __ARGS((tokenInfo *const token, memberInfo *const pMember));
static boolean skipToEnd __ARGS((tokenInfo *const token, memberInfo *const pMember));
static boolean skipEntityType __ARGS((tokenInfo *const token));
static boolean parseFeature __ARGS((const boolean isExported, tokenInfo *const token, memberInfo *const pMember));
static boolean parseExport __ARGS((tokenInfo *const token, boolean *const pIsExported));

/*============================================================================
=   Function definitions
============================================================================*/

extern void buildEiffelKeywordHash()
{
    int lastInitialChar = '\0';
    size_t i;
    const size_t numHashEntries = sizeof(EiffelKeywordHash) /
				  sizeof(EiffelKeywordHash[0]);

    /*  Clear all hash entries.
     */
    for (i = 0  ;  i < numHashEntries  ;  ++i)
	EiffelKeywordHash[i] = -1;

    /*  Set those hash entries corresponding to keywords.
     */
    for (i = 0  ;  i < EiffelKeywordTableSize  ;  ++i)
    {
	const unsigned char initialChar =
				tolower((int)EiffelKeywordTable[i].name[0]);

	if (initialChar != lastInitialChar)
	{
	    EiffelKeywordHash[hashIndex(initialChar)] = i;
	    lastInitialChar = initialChar;
	}
    }
}

static boolean vStringEquiv( string, s )
    vString *const string;
    const char *const s;
{
    return strequiv(string->buffer, s);
}

/*----------------------------------------------------------------------------
*-	Tag generation functions
----------------------------------------------------------------------------*/

static void qualifyClassTag( name, pMember )
    const char *const name;
    memberInfo *const pMember;
{
    tagInfo tag;

    tag.location    = File.seek;
    tag.lineNumber  = File.lineNumber;
    DebugStatement( clearString(tag.name, MaxNameLength); )
    strncpy(tag.name, name, (size_t)MaxNameLength);
    tag.name[MaxNameLength - 1] = '\0';

    pMember->type	= MEMBER_NONE;
    pMember->visibility	= VIS_UNDEFINED;
    pMember->persistent	= FALSE;
    DebugStatement( clearString(pMember->parent, MaxNameLength); )
    pMember->parent[0]	= '\0';
    strncpy(pMember->parent, name, (size_t)MaxNameLength);
    pMember->parent[MaxNameLength - 1] = '\0';

    makeTag(&tag, pMember, SCOPE_GLOBAL, TAG_CLASS);

    /*  Set up the member info for feature tags.
     */
    pMember->type	= MEMBER_CLASS;
    pMember->visibility	= VIS_PUBLIC;
    pMember->persistent	= TRUE;
    strncpy(pMember->parent, name, (size_t)MaxNameLength);
    pMember->parent[MaxNameLength - 1] = '\0';
}

static void qualifyFeatureTag( name, pMember, isExported )
    const char *const name;
    memberInfo *const pMember;
    const boolean isExported;
{
    const tagScope scope = isExported ? SCOPE_GLOBAL : SCOPE_STATIC;
    tagInfo tag;

    tag.location    = File.seek;
    tag.lineNumber  = File.lineNumber;
    DebugStatement( clearString(tag.name, MaxNameLength); )
    strncpy(tag.name, name, (size_t)MaxNameLength);
    tag.name[MaxNameLength - 1] = '\0';

    makeTag(&tag, pMember, scope, TAG_FEATURE);
}

static void qualifyLocalTag( name, pMember )
    const char *const name;
    memberInfo *const pMember;
{
    tagInfo tag;

    tag.location    = File.seek;
    tag.lineNumber  = File.lineNumber;
    DebugStatement( clearString(tag.name, MaxNameLength); )
    strncpy(tag.name, name, (size_t)MaxNameLength);
    tag.name[MaxNameLength - 1] = '\0';

    makeTagEntry(&tag, pMember, SCOPE_STATIC, TAG_LOCAL,
		 (boolean)(Option.locate != EX_PATTERN));
}

/*----------------------------------------------------------------------------
*-	Parsing functions
----------------------------------------------------------------------------*/

static int skipToCharacter( c )
    const int c;
{
    int d;

    do
    {
	d = fileGetc();
    } while (d != EOF  &&  d != c);

    return d;
}

/*  If a numeric is passed in 'c', this is used as the first digit of the
 *  numeric being parsed.
 */
static const char *parseInteger( c )
    int c;
{
    static vString *string = NULL;

    if (string == NULL)
	string = vStringNew();
    vStringClear(string);

    if (c == '-')
    {
	vStringPut(string, c);
	c = fileGetc();
    }
    else if (! isdigit(c))
	c = fileGetc();
    while (c != EOF  &&  (isdigit(c)  ||  c == '_'))
    {
	vStringPut(string, c);
	c = fileGetc();
    }
    vStringPut(string, '\0');
    fileUngetc(c);

    return vStringValue(string);
}

static const char *parseNumeric( c )
    int c;
{
    static vString *string = NULL;

    if (string == NULL)
	string = vStringNew();
    vStringClear(string);
    vStringCat(string, parseInteger(c));

    c = fileGetc();
    if (c == '.')
    {
	vStringPut(string, c);
	vStringCat(string, parseInteger('\0'));
    }

    c = fileGetc();
    if (tolower(c) == 'e')
    {
	vStringPut(string, c);
	vStringCat(string, parseInteger('\0'));
    }
    vStringPut(string, '\0');

    return vStringValue(string);
}

static int parseEscapedCharacter()
{
    int d = '\0';
    int c = fileGetc();

    switch (c)
    {
	case 'A':	d = '@';	break;
	case 'B':	d = '\b';	break;
	case 'C':	d = '^';	break;
	case 'D':	d = '$';	break;
	case 'F':	d = '\f';	break;
	case 'H':	d = '\\';	break;
	case 'L':	d = '~';	break;
	case 'N':	d = '\n';	break;
	case 'Q':	d = '`';	break;
	case 'R':	d = '\r';	break;
	case 'S':	d = '#';	break;
	case 'T':	d = '\t';	break;
	case 'U':	d = '\0';	break;
	case 'V':	d = '|';	break;
	case '%':	d = '%';	break;
	case '\'':	d = '\'';	break;
	case '"':	d = '"';	break;
	case '(':	d = '[';	break;
	case ')':	d = ']';	break;
	case '<':	d = '{';	break;
	case '>':	d = '}';	break;

	case '\n': skipToCharacter('%'); break;

	case '/':
	{
	    const char *value = parseInteger('\0');
	    const unsigned long ascii = atol(value);

	    c = fileGetc();
	    if (c == '/'  &&  ascii < 256)
		d = ascii;
	    break;
	}

	default: break;
    }
    return d;
}

static int parseCharacter()
{
    int c = fileGetc();
    int result = c;

    if (c == '%')
	result = parseEscapedCharacter();

    c = fileGetc();
    if (c != '\'')
	skipToCharacter('\n');

    return result;
}

static void parseString( string )
    vString *const string;
{
    int c = fileGetc();

    while (c != EOF  &&  c != '"')
    {
	if (c == '%')
	    c = parseEscapedCharacter();
	vStringPut(string, c);
	c = fileGetc();
    }
    vStringPut(string, '\0');
}

/*  Read a C identifier beginning with "firstChar" and places it into "name".
 */
static void parseIdentifier( string, firstChar )
    vString *const string;
    const int firstChar;
{
    int c = firstChar;

    do
    {
	vStringPut(string, c);
	c = fileGetc();
    } while (isident(c));

    vStringPut(string, '\0');
    fileUngetc(c);		/* unget non-identifier character */
}

static void parseFreeOperator( string, firstChar )
    vString *const string;
    const int firstChar;
{
    int c = firstChar;

    do
    {
	vStringPut(string, c);
	c = fileGetc();
    } while (c > ' ');

    vStringPut(string, '\0');
    fileUngetc(c);		/* unget non-identifier character */
}

/*  Analyzes the identifier contained in a statement described by the
 *  statement structure and adjusts the structure according the significance
 *  of the identifier.
 */
static keywordId analyzeToken( name )
    const char *const name;
{
    keywordId id = KEYWORD_NONE;
    int initialChar = tolower((int)name[0]);
    const short hash = EiffelKeywordHash[hashIndex(initialChar)];

    if (hash >= 0)
    {
	unsigned int i;

	for (i = hash  ;  i < EiffelKeywordTableSize  ;  ++i)
	{
	    const keywordDesc *pKw = &EiffelKeywordTable[i];

	    if (tolower((int)pKw->name[0]) != initialChar)
		break;

	    if (strequiv(pKw->name, name))
	    {
		id = pKw->id;
		break;
	    }
	}
    }
    return id;
}

static boolean parseToken( token )
    tokenInfo *const token;
{
    boolean more = TRUE;
    int c;

    token->type    = TOKEN_UNDEFINED;
    token->keyword = KEYWORD_NONE;
    vStringClear(token->string);

getNextChar:

    do
	c = fileGetc();
    while (c == '\t'  ||  c == ' '  ||  c == '\n');

    switch (c)
    {
	case EOF:	token->type = TOKEN_EOF; more = FALSE;	break;

	case '!':	token->type = TOKEN_BANG;		break;
	case '$':	token->type = TOKEN_DOLLAR;		break;
	case '(':	token->type = TOKEN_OPEN_PAREN;		break;
	case ')':	token->type = TOKEN_CLOSE_PAREN;	break;
	case ',':	token->type = TOKEN_COMMA;		break;
	case '.':	token->type = TOKEN_DOT;		break;
	case ';':	token->type = TOKEN_SEPARATOR;		break;
	case '[':	token->type = TOKEN_OPEN_BRACKET;	break;
	case ']':	token->type = TOKEN_CLOSE_BRACKET;	break;
	case '{':	token->type = TOKEN_OPEN_BRACE;		break;
	case '}':	token->type = TOKEN_CLOSE_BRACE;	break;

	case '+':
	case '*':
	case '^':
	case '=':	token->type = TOKEN_OPERATOR;		break;

	case '-':
	    c = fileGetc();
	    if (c != '-')		/* is this the start of a comment? */
		fileUngetc(c);
	    else
	    {
		skipToCharacter('\n');
		goto getNextChar;
	    }
	    break;

	case '?':
	case ':':
	    c = fileGetc();
	    if (c == '=')
		token->type = TOKEN_OPERATOR;
	    else
	    {
		token->type = TOKEN_COLON;
		fileUngetc(c);
	    }
	    break;

	case '<':
	    c = fileGetc();
	    if (c != '='  &&  c != '>')
		fileUngetc(c);
	    token->type = TOKEN_OPERATOR;
	    break;

	case '>':
	    c = fileGetc();
	    if (c != '='  &&  c != '>')
		fileUngetc(c);
	    token->type = TOKEN_OPERATOR;
	    break;

	case '/':
	    c = fileGetc();
	    if (c != '/'  &&  c != '=')
		fileUngetc(c);
	    token->type = TOKEN_OPERATOR;
	    break;

	case '\\':
	    c = fileGetc();
	    if (c != '\\')
		fileUngetc(c);
	    token->type = TOKEN_OPERATOR;
	    break;

	case '"':
	    token->type = TOKEN_STRING;
	    parseString(token->string);
	    break;

	case '\'':
	    token->type = TOKEN_CHARACTER;
	    parseCharacter();
	    break;

	default:
	    if (isalpha(c))
	    {
		parseIdentifier(token->string, c);
		token->keyword = analyzeToken(vStringValue(token->string));
		if (token->keyword == KEYWORD_NONE)
		    token->type = TOKEN_IDENTIFIER;
		else
		    token->type = TOKEN_KEYWORD;
	    }
	    else if (isdigit(c))
	    {
		vStringCat(token->string, parseNumeric(c));
		token->type = TOKEN_NUMERIC;
	    }
	    else if (isFreeOperatorChar(c))
	    {
		parseFreeOperator(token->string, c);
		token->type = TOKEN_OPERATOR;
	    }
	    else
	    {
		token->type = TOKEN_UNDEFINED;
		Assert(token->type != TOKEN_UNDEFINED);
	    }
	    break;
    }
    return more;
}

/*----------------------------------------------------------------------------
*-	Scanning functions
----------------------------------------------------------------------------*/

static boolean isIdentifierMatch( token, name )
    const tokenInfo *const token;
    const char *const name;
{
    return (boolean)(token->type == TOKEN_IDENTIFIER  &&
		     vStringEquiv(token->string, name));
}

static boolean locateToken( type, skipPast, token )
    const tokenType type;
    const boolean skipPast;
    tokenInfo *const token;
{
    boolean ok;

    do
	ok = parseToken(token);
    while (ok  &&  token->type != type);

    if (ok  &&  skipPast)
	ok = parseToken(token);

    return ok;
}

static boolean locateKeyword( keyword, skipPast, token )
    const keywordId keyword;
    const boolean skipPast;
    tokenInfo *const token;
{
    boolean ok;

    do
	ok = parseToken(token);
    while (ok  &&  token->keyword != keyword);

    if (ok  &&  skipPast)
	ok = parseToken(token);

    return ok;
}

static boolean parseLocal( token, pMember )
    tokenInfo *const token;
    memberInfo *const pMember;
{
    boolean ok;

    do
    {
	if (token->type == TOKEN_IDENTIFIER)
	    qualifyLocalTag (vStringValue(token->string), pMember);

	ok = parseToken(token);
	if (ok  &&  token->type == TOKEN_COLON)
	    ok = skipEntityType(token);
    } while (ok &&  token->type != KEYWORD_do  &&  token->type != KEYWORD_once);

    return ok;
}

static boolean skipToEnd( token, pMember )
    tokenInfo *const token;
    memberInfo *const pMember;
{
    boolean ok = TRUE;
    int depth = 1;

    while (ok  &&  depth > 0)
    {
	ok = parseToken(token);
	if (ok) switch (token->keyword)
	{
	    case KEYWORD_check:
	    case KEYWORD_debug:
	    case KEYWORD_from:
	    case KEYWORD_if:
	    case KEYWORD_inspect:
		++depth;
		break;

	    case KEYWORD_local:
		ok = parseToken(token);
		if (ok)
		    parseLocal(token, pMember);
		break;

	    case KEYWORD_end:
		--depth;
		break;

	    default:
		break;
	}
    }

    /*  Skip to token following "end".
     */
    if (ok)
	ok = parseToken(token);

    return ok;
}

static boolean skipEntityType( token )
    tokenInfo *const token;
{
    boolean ok;

    /*  When called, we are at the ':'. Move to the type.
     */
    ok = parseToken(token);

    /*  Skip over the type name, with possible generic parameters.
     *  */
    if (ok  &&  token->keyword == KEYWORD_like)
	ok = parseToken(token);
    if (ok)
	ok = parseToken(token);
    if (ok  &&  token->type == TOKEN_OPEN_BRACKET)
	ok = locateToken(TOKEN_CLOSE_BRACKET, TRUE, token);

    return ok;
}

static boolean parseFeature( isExported, token, pMember )
    const boolean isExported;
    tokenInfo *const token;
    memberInfo *const pMember;
{
    boolean ok = TRUE;

    /*  Generate tag for either the feature identifier, or the string
     *  following either prefix or infix.
     */
    if (token->type == TOKEN_IDENTIFIER)
	qualifyFeatureTag (vStringValue(token->string), pMember, isExported);
    else if (token->keyword == KEYWORD_infix  ||
	     token->keyword == KEYWORD_prefix)
    {
	ok = parseToken(token);
	if (token->type == TOKEN_STRING)
	    qualifyFeatureTag (vStringValue(token->string), pMember,
			       isExported);
    }

    if (ok)
	ok = parseToken(token);

    /*  Skip over arguments and return type.
     */
    if (ok  &&  token->type == TOKEN_OPEN_PAREN)
	ok = locateToken(TOKEN_CLOSE_PAREN, TRUE, token);
    if (ok  &&  token->type == TOKEN_COLON)
	ok = skipEntityType(token);

    /*  Skip over body if one exists.
     */
    if (ok  &&  token->keyword == KEYWORD_is)
	ok = skipToEnd(token, pMember);

    return ok;
}

static boolean parseExport( token, pIsExported )
    boolean *const pIsExported;
    tokenInfo *const token;
{
    boolean ok;

    *pIsExported = TRUE;

    ok = parseToken(token);
    if (ok   &&  token->type == TOKEN_OPEN_BRACE)
    {
	ok = parseToken(token);
	if (ok)
	{
	    if (token->type == TOKEN_CLOSE_BRACE)
	    {
		*pIsExported = FALSE;
		ok = parseToken(token);
	    }
	    else
	    {
		if (isIdentifierMatch(token, "NONE"))
		    *pIsExported = FALSE;
		ok = locateToken(TOKEN_CLOSE_BRACE, TRUE, token);
	    }
	}
    }
    return ok;
}

extern boolean createEiffelTags()
{
    boolean isExported = TRUE;
    tokenInfo token;
    memberInfo member;

    token.string = vStringNew();

    if (locateKeyword(KEYWORD_class, TRUE, &token))
    {
	if (token.type == TOKEN_IDENTIFIER)
	    qualifyClassTag(vStringValue(token.string), &member);

	if (locateKeyword(KEYWORD_feature, FALSE, &token))
	{
	    boolean ok = TRUE;

	    do
	    {
		if (token.keyword == KEYWORD_invariant)
		    ok = locateKeyword(KEYWORD_feature, FALSE, &token);
		if (ok  &&  token.keyword == KEYWORD_feature)
		    ok = parseExport(&token, &isExported);
		if (ok)
		    ok = parseFeature(isExported, &token, &member);
	    } while (ok  &&  token.keyword != KEYWORD_end);
	}
    }
    return TRUE;
}

/* vi:set tabstop=8 shiftwidth=4: */
