/*****************************************************************************
*   $Id: read.c,v 6.14 1998/10/18 15:37:35 darren Exp $
*
*   Copyright (c) 1996-1998, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains low level source and tag file read functions (line
*   splicing and newline conversion for source files are performed at this
*   level).
*****************************************************************************/

/*============================================================================
=   Include files
============================================================================*/
#include "general.h"

#include <string.h>

#include "ctags.h"
#define FILE_WRITE
#include "read.h"

#include "main.h"
#include "options.h"
#include "debug.h"

/*============================================================================
=   Data declarations
============================================================================*/

typedef struct _langTab {
    const char *extension;
    langType language;
} langTab;

/*============================================================================
=   Data definitions
============================================================================*/

inputFile File = {
    NULL,		/* name of input file */
    NULL,		/* path of input file */
    NULL,		/* file pointer */
    0L,			/* line number of the input file */
    -1L,		/* current seek location in file */
    '\0',		/* ungotten character */
    FALSE,		/* was previous character a newline? */
    FALSE,		/* is input file a header file? */
    LANG_IGNORE,	/* language of current file */
    {
	NULL,		/* name to report as source file */
	0L,		/* line number of the source file */
	FALSE,		/* is input file a header file? */
	LANG_IGNORE	/* language of current file */
    }
};

/*============================================================================
=   Function prototypes
============================================================================*/
static void setSourceFileName __ARGS((const char *const fileName));
static void fileNewline __ARGS((void));

/*============================================================================
=   Function definitions
============================================================================*/

/*----------------------------------------------------------------------------
 *	Generic line reading with automatic buffer sizing
 *--------------------------------------------------------------------------*/

extern char *readLine( vLine, fp )
    vString *const vLine;
    FILE *const fp;
{
    char *line = NULL;

    vStringClear(vLine);
    if (fp == NULL)		/* to free memory allocated to buffer */
	error(FATAL, "NULL file pointer");
    else
    {
	boolean reReadLine;

	/*  If reading the line places any character other than a null or a
	 *  newline at the last character position in the buffer (one less
	 *  than the buffer size), then we must resize the buffer and
	 *  reattempt to read the line.
	 */
	do 
	{
	    const long startOfLine = ftell(fp);
	    char *const pLastChar = vStringValue(vLine) + vStringSize(vLine) -2;

	    reReadLine = FALSE;
	    *pLastChar = '\0';
	    line = fgets(vStringValue(vLine), (int)vStringSize(vLine), fp);
	    if (line == NULL)
	    {
		if (! feof(fp))
		    error(FATAL | PERROR, "Failure on attempt to read file");
	    }
	    else
	    {
		/*  Check for buffer overflow.
		 */
		if (*pLastChar != '\0'  &&  *pLastChar != '\n')
		{
		    if ((reReadLine = vStringAutoResize(vLine)))
			fseek(fp, startOfLine, SEEK_SET);
		    else
			error(FATAL | PERROR,
			      "input line too big; out of memory");
		}
	    }
	} while (reReadLine);
    }
    return line;
}

/*----------------------------------------------------------------------------
 *	Source file access functions
 *--------------------------------------------------------------------------*/

static void setSourceFileName( fileName )
    const char *const fileName;
{
    if (File.source.name == NULL)
	File.source.name = allocFilenameBuffer(NULL);

    strcpy(File.source.name, fileName);
    if (strlen(fileName) > TagFile.max.file)
	TagFile.max.file = strlen(fileName);
}

extern void setSourceFileParameters( fileName, language, isHeader )
    const char *const fileName;
    const langType language;
    const boolean isHeader;
{
    if (isAbsolutePath(fileName)  ||  File.path[0] == '\0')
	setSourceFileName(fileName);
    else
    {
	char *const pathName = allocFilenameBuffer(NULL);

	combinePathAndFile(pathName, File.path, fileName);
	setSourceFileName(pathName);
	free(pathName);
    }
    File.source.isHeader = isHeader;
    File.source.language = language;
}

extern void setSourceFileLine( lineNumber )
    const unsigned long lineNumber;
{
    File.source.lineNumber = lineNumber;
}

/*  This function opens a source file, and resets the line counter.  If it
 *  fails, it will display an error message and leave the File.fp set to NULL.
 */
extern boolean fileOpen( fileName, language, isHeader )
    const char *const fileName;
    const langType language;
    const boolean isHeader;
{
    boolean opened = FALSE;

    /*	If another file was already open, then close it.
     */
    if (File.fp != NULL)
    {
	fclose(File.fp);		/* close any open source file */
	File.fp = NULL;
    }

    File.fp = fopen(fileName, "rb");  /* must be binary mode for fseek() */
    if (File.fp == NULL)
	error(WARNING | PERROR, "cannot open \"%s\"", fileName);
    else
    {
	const char *const tail = baseFilename(fileName);
	opened = TRUE;

	if (File.name == NULL) File.name = allocFilenameBuffer(NULL);
	if (File.path == NULL) File.path = allocFilenameBuffer(NULL);
	strcpy(File.name, fileName);
	if (tail == fileName)
	    File.path[0] = '\0';	    /* no path specified */
	else
	{
	    const size_t length = tail - fileName - 1;

	    strncpy(File.path, fileName, length);
	    File.path[length] = '\0';
	}
	File.lineNumber = 0L;
	File.seek	= 0L;
	File.newLine	= TRUE;
	File.isHeader	= isHeader;
	File.language	= language;

	setSourceFileName(fileName);
	File.source.lineNumber = 0L;
	File.source.isHeader = isHeader;
	File.source.language = language;

	DebugStatement( debugOpen(fileName, File.isHeader, File.language); )
    }
    return opened;
}

extern void fileClose()
{
    if (File.fp != NULL)
    {
	/*  The line count of the file is 1 too big, since it is one-based
	 *  and is incremented upon each newline.
	 */
	if (Option.printTotals)
	    addTotals(0, File.lineNumber - 1L, getFileSize(File.name));

	fclose(File.fp);
	File.fp = NULL;
    }
}

/*  Action to take for each encountered source newline.
 */
static void fileNewline()
{
    File.newLine = FALSE;
    File.seek	 = ftell(File.fp) - 1L;	/* already read 1st character of line */
    File.lineNumber++;
    File.source.lineNumber++;
    DebugStatement( if (Option.breakLine == File.lineNumber) lineBreak(); )
    DebugStatement( debugPrintf(DEBUG_RAW, "%6d: ", File.lineNumber); )
}

/*  This function reads a single character from the stream. 
 */
extern int fileGetc()
{
    int	c;

    /*	If there is an ungotten character, then return it.  Don't do any
     *	other processing on it, though, because we already did that the
     *	first time it was read.
     */
    if (File.ungetch != '\0')
    {
	c = File.ungetch;
	File.ungetch = '\0';
	return c;	    /* return here to avoid re-calling debugPutc() */
    }

nextChar:	/* not structured, but faster for this critical path */
    c = getc(File.fp);

    /*	If previous character was a newline, then we're starting a line.
     */
    if (File.newLine  &&  c != EOF)
	fileNewline();

    if (c == NEWLINE)
	File.newLine = TRUE;
    else if (c == CRETURN)
    {
	/*  Turn line breaks into a canonical form. The three commonly
	 *  used forms if line breaks: LF (UNIX), CR (MacIntosh), and
	 *  CR-LF (MS-DOS) are converted into a generic newline.
	 */
	const int next = getc(File.fp);		/* is CR followed by LF? */

	if (next != NEWLINE)
	    ungetc(next, File.fp);

	c = NEWLINE;				/* convert CR into newline */
	File.newLine = TRUE;
    }
    else if (c == BACKSLASH)
    {
	int next = getc(File.fp);

	if (next == CRETURN)
	{
	    next = getc(File.fp);		/* is CR followed by LF? */
	    if (next != NEWLINE)
		ungetc(next, File.fp);
	    next = NEWLINE;			/* convert CR into newline */
	}
	if (next == NEWLINE)
	{
	    File.newLine = TRUE;
	    DebugStatement( debugPutc(BACKSLASH, DEBUG_RAW);
			    debugPutc(NEWLINE, DEBUG_RAW);)
	    goto nextChar;			/* through away "\NEWLINE" */
	}
	else
	    ungetc(next, File.fp);	/* put back character after BACKSLASH */
    }
    DebugStatement( debugPutc(c, DEBUG_RAW); )
    return c;
}

extern void fileUngetc( c )
    int c;
{
    File.ungetch = c;
}

/*  Places into the line buffer the contents of the line referenced by
 *  "location".
 */
extern char *readSourceLine( vLine, location )
    vString *const vLine;
    const long location;
{
    const long orignalPosition = ftell(File.fp);
    char *line;

    fseek(File.fp, location, SEEK_SET);
    line = readLine(vLine, File.fp);	
    if (line == NULL)
	error(FATAL, "Unexpected end of file: %s", File.name);
    fseek(File.fp, orignalPosition, SEEK_SET);

    return line;
}

/* vi:set tabstop=8 shiftwidth=4: */
